#%% ###################################################################### Imports ######################################################################
import os # For cwd and chdir (file oeprations), required to find AdaptiveFilters, might generate errors in some IDEs
PATH = "C:/Users/Stéphane/BCP, A/00 TU/Digital Signal Processing/00 Modules/Adaptive Filter/my project/Code"
os.chdir(PATH) # set current python directory to PATH

import AdaptiveFilters as AF # file containing the implemented classes

import pickle # to store all computations

# Numpy
import numpy as np
#import numpy.linalg as npla
# np.set_printoptions(threshold=np.inf) # print entire arrays
np.set_printoptions(linewidth=np.inf) # don't auto-wrap arrays

# Matplotlib
import matplotlib.pyplot as plt
#plt.rcParams['figure.figsize'] = [6, 6]
#plt.rcParams['figure.dpi'] = 600

from scipy.io import wavfile, loadmat # read and write wave/matlab files (24bit not supported for wav)
from scipy.signal import lfilter, firwin, freqz, dlti, dlsim # for signal filtering

##################################################################### Data Management #####################################################################
PATH += "/00 Data"
SigLen = 100 # all provided signal are 10k long, can be shortened to improve simulation speed
Data = {} # declare Data to be a dictionary

for System in os.listdir(PATH): # iterate over entire folder
   Temp = loadmat(PATH + "/" + System) # load all files in given directory
   del Temp['__header__'], Temp['__version__'], Temp['__globals__'] # delete Matlab formatting stuff no-one cares about
   Data[System.split(".")[0]] = Temp # create dict entry, strip ".mat" in the name and put cleaned data as subdict

Data["KLMS"] = {"D_": Data['KLMS_Test']['x_test'], "X":Data['KLMS_Training']['x_training']} # Keep same Data structure as before
del Data['KLMS_Training'], Data['KLMS_Test'] # delete original entries


for System in list(Data): # iterate over entire data set (the cutting operations are ignored by shorter lengths (KLMS-Data)
   Data[System]["D_"] = Data[System]['D_'].ravel()[:SigLen]#/ np.max(np.abs(Data[System]['D_'].ravel()[:SigLen])) # somehow 4 samples too long compared to X, normalize
   Data[System]["X"] = Data[System]['X'].ravel()[:SigLen]#/ np.max(np.abs(Data[System]['X'].ravel()[:SigLen])) # shorten and normalize

# Add known Systems for comparison
Data["FIR"]["LTI"] = dlti([0.7, 0.1, -0.03, 0.18, -0.24], [1.0]+[0.0]*4, dt=1) # denomiantor is 1 for FIRs, padded with 0's
Data["IIR"]["LTI"] = dlti([1.0, 0.0], [0.82, -0.03], dt=1) # numerator = 1, being a pure IIR, the CTor normalizes the filter coeffs to a_0=1

#Fig, Ax = plt.subplots() # create figure
#plt.plot(Data["IIR"]["LTI"].impulse()[1][0]) # plot IIR system impulse responce
#AF.PltFormat(Fig, Ax, "Impulse response", "Time in Samples", [], "Amplitude", [], [], Show = False)

############################################################## Exercice 1-4: FIR+IIR #####################################################################
AdaptResults = {"LMS":{}, "pRLS":{}} # declare dict to store results for both algorithms, both containing a noiseless (NL) subdict

## uncomment one of the filters to apply to this section, the rest of the code adapts automatically.
## the system change will be recognized with if(Filter[-1] == "l") as both end by an 'l'.

#Filter = "FIR" # Exercise 1
Filter = "IIR" # Exercise 2 
#Filter = "FIR_SysWechsel" # Exercise 3
#Filter = "IIR_SysWechsel" # Exercise 4


## *********************************************************************** LMS *********************************************************************
# -------------------------------------------------------- 1. Generalized Gridsearches to confirm theory -----------------------------------------------
# this will genarate a lot of overflow warnign and NaN but this is part of the testing procedure
if (Filter in ["FIR", "IIR"]): # this procedure is only necessary for the first tests
   for NTaps in [1,2,5]: # 1,2,5
      AdaptResults["LMS"]["NTaps" + str(NTaps)] = {} # create common dict for following for loop

      for NV in [0.001, 0.1, 1, 10]: # noise variances: 0.001, 0.1, 1, 10

         Input = np.random.normal(0, NV, SigLen) + Data[Filter]["X"] # zero mean NV-variance noise addition
         MaxMu, InputPower = AF.LMSconverge(Input, NTaps=NTaps, Mu=None, Print=False)[2:] # get conversion predictions and excludes bool and time
         MuList = [1, 0.5, 0.25, 0.1, 0.05, 0.025, 0.01, 0.005, 0.0025, 0.001, 0.0005]

         AdaptResults["LMS"]["NTaps" + str(NTaps)]["N_" + str(NV)] = {} # create current subdict
         ThisDict = AdaptResults["LMS"]["NTaps" + str(NTaps)]["N_" + str(NV)] # ThisDict (pointer) to avoid long name
         print("\nCurrently in NV =", NV, "with Power:", np.round(InputPower,4),"and maximal converging mu:", MaxMu, "\n")

         ThisDict["BestPar"], ThisDict["Tab"] = AF.gridsearch({"FilterType":["LMS"], "NTaps": [NTaps], "Mu":MuList}, Input, Data[Filter]["D_"], Plot=False)
         AF.LMS(**ThisDict["BestPar"]).adapt(Input, Data[Filter]["D_"], Plot = True) # adapt and plot best Filter


##--------------------------------------------------------- 2. Filter order vs Step size and MSE (no noise) -------------------------------------------
AdaptResults["LMS"]["OrderVsStepsize"] = {}
Fig, Ax = plt.subplots() # create figure
if(Filter[-1] == "l"): MaxMSE = 0 # required for the hight of the separation bar
TapList = [1, 2, 5]

for NTaps in TapList:

   AdaptResults["LMS"]["OrderVsStepsize"][str(NTaps) + "Taps"] = {} # create current subdict
   ThisDict = AdaptResults["LMS"]["OrderVsStepsize"][str(NTaps) + "Taps"] # Alias for brevity

   Input = Data[Filter]["X"]
   MaxMu = AF.LMSconverge(Input, NTaps=NTaps, Mu=None, Print=False)[2] # get MaxMu only, no predictions
   MuList = list(np.logspace(np.log2(MaxMu/2), np.log2(MaxMu/2048), num=60, base=2.0)) # range working well for all noise intensities

   ThisDict["BestPar"], ThisDict["Table"] = AF.gridsearch({"FilterType":["LMS"], "NTaps": [NTaps], "Mu":MuList}, Input, Data[Filter]["D_"], Plot=False)
   ThisDict["MSE"], ThisDict["Weights"] = AF.LMS(**ThisDict["BestPar"]).adapt(Input, Data[Filter]["D_"], Plot = True)[2:] # adapt and plot best Filter, keep MSE only

   if(NTaps == 1): LegendString = str(NTaps) + " Tap   with µ = " + '{:.3e}'.format(ThisDict["BestPar"]["Mu"]) # put Mu in scientific notation
   else: LegendString = str(NTaps) + " Taps with µ = " + '{:.3e}'.format(ThisDict["BestPar"]["Mu"]) # put Mu in scientific notation
   Ax.plot(ThisDict["MSE"], label=LegendString)
   if(Filter[-1] == "l" and (ThisDict["MSE"].max() > MaxMSE) ): MaxMSE = ThisDict["MSE"].max() # update bar height if Syschange

AF.PltFormat(Fig, Ax, "", "Time in Samples", [0, SigLen], "Mean Squared Error", [], [], Show = False)
Ax.legend(loc="best")# print the legend created during the loop
if(Filter[-1] == "l"): Ax.vlines(5000,0, MaxMSE) # make the vertical bar

for NTaps in TapList:
   print("LMS and NV=0 with", NTaps, "taps yields following weights:", AdaptResults["LMS"]["OrderVsStepsize"][str(NTaps) + "Taps"]["Weights"],
         "with MSE=", np.round(AdaptResults["LMS"]["OrderVsStepsize"][str(NTaps) + "Taps"]["MSE"][-1], 5))


## ---------------------------------------------- 3. NTap = 5 vs noise level but optimal hyper-parameters  -------------------------------------------
AdaptResults["LMS"]["NoiseResistance"] = {}
Fig, Ax = plt.subplots() # create figure
if(Filter[-1] == "l"): MaxMSE = 0 # required for the hight of the separation bar

Taps = 5 # change to 1 or 2 for different plots (but all similar to 5 taps)
NVList = [0.001, 0.1, 1, 10] #0.001, 0.1, 1, 10
for NV in NVList: # noise variances

   AdaptResults["LMS"]["NoiseResistance"]["N_" + str(NV)] = {} # create current subdict
   ThisDict = AdaptResults["LMS"]["NoiseResistance"]["N_" + str(NV)] # Alias for brevity

   Input = np.random.normal(0, NV, SigLen) + Data[Filter]["X"] # zero mean NV-variance noise addition
   MaxMu = AF.LMSconverge(Input, NTaps=5, Mu=None, Print=False)[2] # get MaxMu only, no predictions
   MuList = list(np.logspace(np.log2(MaxMu/2), np.log2(MaxMu/2048), num=50, base=2.0)) # range working well for all noise intensities

   ThisDict["BestPar"], ThisDict["Table"] = AF.gridsearch({"FilterType":["LMS"], "NTaps": [Taps], "Mu":MuList}, Input, Data[Filter]["D_"], Plot=False)
   ThisDict["MSE"], ThisDict["Weights"] = AF.LMS(**ThisDict["BestPar"]).adapt(Input, Data[Filter]["D_"], Plot = True)[2:] # adapt and plot best Filter, keep MSE + weights only

   LegendString = "Noise Amplitude = " + str(NV) + "; with Mu = " + '{:.3e}'.format(ThisDict["BestPar"]["Mu"]) # put Mu in scientific notation
   Ax.plot(ThisDict["MSE"], label=LegendString)
   if(Filter[-1] == "l" and (ThisDict["MSE"].max() > MaxMSE) ): MaxMSE = ThisDict["MSE"].max() # update bar height if Syschange

AF.PltFormat(Fig, Ax, str(Filter) + " LMS with " + str(Taps) + "Taps", "Time in Samples", [-20, SigLen], "Mean Squared Error", [], [], Show = False)
Ax.legend(loc="best")# print the legend created during the loop
if(Filter[-1] == "l"): Ax.vlines(5000,0, MaxMSE) # make the vertical bar

for NV in NVList:
   print(Taps, "Tap(s) LMS and NV =", NV, "yields following weights:", np.round(AdaptResults["LMS"]["NoiseResistance"]["N_" + str(NV)]["Weights"], 6),
         "with MSE=",np.round(AdaptResults["LMS"]["NoiseResistance"]["N_" + str(NV)]["MSE"][-1], 5))


# ************************************************************************* pRLS ***********************************************************************

## ------------------------------------------------------------- 1. Matrix init -----------------------------------------------
# procedure done for all tests (FIR and IIR only take p=1, then search for the best init)

TestOnlyp = True # Change accordingly to different test
AdaptResults["pRLS"]["p_MatrixInit"] ={} # create dict for the following for loop

for NTaps in [5]: # 
   AdaptResults["pRLS"]["p_MatrixInit"]["NTaps" + str(NTaps)] = {} # create common dict for following for loop

   for NV in [0.001, 0.1, 1, 10]: # noise variances: 

      Input = np.random.normal(0, NV, SigLen) + Data[Filter]["X"] # zero mean NV-variance noise addition
      if(TestOnlyp): MatrixInitList = [10000]
      else: MatrixInitList = list(np.logspace(np.log2(0.000001), np.log2(100000), num=50, base=2.0)) # range working well for all noise 
      
      pList = list(np.linspace(1.0, 0.9, 20)) # non-stationary environments require to search the best p
      

      AdaptResults["pRLS"]["p_MatrixInit"]["NTaps" + str(NTaps)]["N_" + str(NV)] = {} # create current subdict
      ThisDict = AdaptResults["pRLS"]["p_MatrixInit"]["NTaps" + str(NTaps)]["N_" + str(NV)] # ThisDict (pointer) to avoid long name
      print("\nCurrently in NV =", NV, "with Power:", np.round(np.mean(Input**2), 4))

      ThisDict["BestPar"], ThisDict["Tab"] = AF.gridsearch({"FilterType":["pRLS"], "NTaps": [NTaps], "p": pList,"MatrixInit": MatrixInitList},
                                                           Input, Data[Filter]["D_"], Plot=True)
      # AF.pRLS(**ThisDict["BestPar"]).adapt(Input, Data[Filter]["D_"], Plot = False) # adapt and plot best Filter


#--------------------------------------------------------- 2. Filter order vs MSE (no noise) -------------------------------------------
AdaptResults["pRLS"]["OrderVsMSE"] = {}
Fig, Ax = plt.subplots() # create figure
if(Filter[-1] == "l"): MaxMSE = 0 # required for the hight of the separation bar
TapList = [1,2,5]

for NTaps in TapList:

   AdaptResults["pRLS"]["OrderVsMSE"][str(NTaps) + "Taps"] = {} # create current subdict
   ThisDict = AdaptResults["pRLS"]["OrderVsMSE"][str(NTaps) + "Taps"] # Alias for brevity

   # adapt and plot Filter, keep MSE and weights, M-Init=10000 since no noise, thus best hyper parameter
   Input = Data[Filter]["X"] # no noise addition
   if(Filter[-1] == "l"): pList = list(np.linspace(1.0, 0.90, 40)) # p-values to test
   else: pList = [1] # best value for stationary systems

   ThisDict["BestPar"], ThisDict["Table"] = AF.gridsearch({"FilterType":["pRLS"], "NTaps": [NTaps], "p": pList, "MatrixInit": [10000]},
                                                          Input, Data[Filter]["D_"], Plot=True)
   ThisDict["MSE"], ThisDict["Weights"] = AF.pRLS(**ThisDict["BestPar"]).adapt(Input, Data[Filter]["D_"], Plot = False)[2:] # adapt and plot best Filter, keep MSE + weights only

   if(NTaps == 1): LegendString = str(NTaps) + " Tap   with MI=10k and p=" + str(np.round(ThisDict["BestPar"]["p"],3)) # put p in scientific notation
   else: LegendString = str(NTaps) + " Taps with MI=10k and p=" + str(np.round(ThisDict["BestPar"]["p"],4)) # put p in scientific notation
   Ax.plot(ThisDict["MSE"], label=LegendString)
   if(Filter[-1] == "l" and (ThisDict["MSE"].max() > MaxMSE) ): MaxMSE = ThisDict["MSE"].max() # update bar height if Syschange

AF.PltFormat(Fig, Ax, "", "Time in Samples", [0, SigLen], "Mean Squared Error", [], [], Show = False)
Ax.legend(loc="best")# print the legend created during the loop
if(Filter[-1] == "l"): Ax.vlines(5000,0, MaxMSE) # make the vertical bar

for NTaps in TapList:
   print("pRLS and NV=0 with", NTaps, "taps yields following weights:", AdaptResults["pRLS"]["OrderVsMSE"][str(NTaps) + "Taps"]["Weights"],
         "with MSE=",np.round(AdaptResults["pRLS"]["OrderVsMSE"][str(NTaps) + "Taps"]["MSE"][-1], 5))


# ---------------------------------------------- 3. NTap = 5 vs noise level but optimal hyper-parameters  -------------------------------------------
AdaptResults["pRLS"]["NoiseResistance"] = {}
Fig, Ax = plt.subplots() # create figure
if(Filter[-1] == "l"): MaxMSE = 0 # required for the hight of the separation bar
NVList = [0.001, 0.1, 1, 10]

Taps = 5 # change to 1 or 2 for different plots (but all similar to 5 taps)
for NV in NVList: # noise variances

   AdaptResults["pRLS"]["NoiseResistance"]["N_" + str(NV)] = {} # create current subdict
   ThisDict = AdaptResults["pRLS"]["NoiseResistance"]["N_" + str(NV)] # Alias for brevity

   Input = np.random.normal(0, NV, SigLen) + Data[Filter]["X"] # zero mean NV-variance noise addition
   MatrixInitList = list(np.logspace(np.log2(0.000001), np.log2(100000), num=60, base=2.0)) # range working well for all noise
   if(Filter[-1] == "l"): pList = list(np.linspace(1.0, 0.94, 20)) # p-values to test
   else: pList = [1] # best value for stationary systems

   ThisDict["BestPar"], ThisDict["Table"] = AF.gridsearch({"FilterType":["pRLS"], "NTaps": [Taps], "p": pList, "MatrixInit": MatrixInitList},
                                                          Input, Data[Filter]["D_"], Plot=False)
   ThisDict["MSE"], ThisDict["Weights"] = AF.pRLS(**ThisDict["BestPar"]).adapt(Input, Data[Filter]["D_"], Plot = True)[2:] # adapt and plot best Filter, keep MSE + weights only

   LegendString = "Noise Amplitude = " + str(NV) + "; MI = " + '{:.3e}'.format(ThisDict["BestPar"]["MatrixInit"]) + "; p=" + str(np.round(ThisDict["BestPar"]["p"],3)) # put p in scientific notation
   Ax.plot(ThisDict["MSE"], label=LegendString)
   if(Filter[-1] == "l" and (ThisDict["MSE"].max() > MaxMSE) ): MaxMSE = ThisDict["MSE"].max() # update bar height if Syschange

AF.PltFormat(Fig, Ax, str(Filter) + " pRLS with " + str(Taps) + "Taps", "Time in Samples", [0, SigLen], "Mean Squared Error", [], [], Show = False)
Ax.legend(loc="best")# print the legend created during the loop
if(Filter[-1] == "l"): Ax.vlines(5000,0, MaxMSE) # make the vertical bar

for NV in NVList:
   print(Taps, "Tap(s) pRLS and NV =", NV, "yields following weights:", np.round(AdaptResults["pRLS"]["NoiseResistance"]["N_" + str(NV)]["Weights"], 5),
         "MSE=", np.round(AdaptResults["pRLS"]["NoiseResistance"]["N_" + str(NV)]["MSE"][-1], 5))


## ************************************************************************* pRLS vs LMS ***********************************************************************
# store and load the correct data
Taps = 5
File = "C:/Users/Acer/BCP, A/00 TU/Digital Signal Processing/00 Modules/Adaptive Filter/my project/Code/" + Filter + str(Taps) + "TData.pkl"

#with open(File, "wb+") as F: # store current dict content
#   pickle.dump(AdaptResults, F)
#   F.close()

#with open(File, 'rb') as f: AdaptResults = pickle.load(f) # load current dict

Fig, Ax = plt.subplots() # create figure
if(Filter[-1] == "l"): MaxMSE = 0 # required for the hight of the separation bar
for NV in [0.001, 0.1, 1, 10]: # noise variances

   LMSDict = AdaptResults["LMS"]["NoiseResistance"]["N_" + str(NV)] # Alias for brevity
   plt.plot(LMSDict["MSE"], label = "LMS:  NA = " + str(NV) + "; with Mu = " + '{:.3e}'.format(LMSDict["BestPar"]["Mu"])) #plot MSE
   print(Taps, "Taps LMS and NV =", NV, "yields following weights:", LMSDict["Weights"]) # display weights

   pRLSDict = AdaptResults["pRLS"]["NoiseResistance"]["N_" + str(NV)] # Alias for brevity
   pRLSLabel = "pRLS: NA = " + str(NV) + "; with MI = " + '{:.3e}'.format(pRLSDict["BestPar"]["MatrixInit"]) +"; p = "+ '{:.3e}'.format(pRLSDict["BestPar"]["p"])
   plt.plot(pRLSDict["MSE"], label = pRLSLabel) #plot MSE
   print(Taps, "Tap(s) pRLS and NV =", NV, "yields following weights:", pRLSDict["Weights"])# display weights
   
   if(Filter[-1] == "l"): # update bar height if Syschange, accounts for both dictionaries
     if(LMSDict["MSE"].max() > MaxMSE ): MaxMSE = LMSDict["MSE"].max() 
     if(pRLSDict["MSE"].max() >MaxMSE ): MaxMSE = pRLSDict["MSE"].max()

AF.PltFormat(Fig, Ax, str(Filter) + " LMS vs pRLS with " + str(Taps) + "Taps", "Time in Samples", [-15, SigLen], "Mean Squared Error", [], [], Show = False)
Ax.legend(loc="best")# print the legend created during the loop
if(Filter[-1] == "l"): Ax.vlines(5000,0, MaxMSE) # make the vertical bar





#%% ############################################################################ Prediction ##############################################################
PredictResuts = {} # general data structure with pre-made entries
TrainData = Data["KLMS"]["X"] # Alias for brevity
TestData = Data["KLMS"]["D_"] # Alias for brevity

for Taps in [5, 10]: # taps for LMS and windowsize for KLMS

   # Dict creation
   PredictResuts[str(Taps) + "Taps"] = {"LMS":{}, "KLMS":{}} # directly create nested subdicts
   LMSDict = PredictResuts[str(Taps) + "Taps"]["LMS"] # Alias for brevity
   KLMSDict = PredictResuts[str(Taps) + "Taps"]["KLMS"] # Alias for brevity


   # LMS
   MaxMu = AF.LMSconverge(TrainData, Taps)[2] # only maximal Mu
   MuList = list(np.logspace(np.log2(MaxMu), np.log2(MaxMu/10), num=40, base=2.0)) # range working well for this data

   LMSDict["BestPar"], LMSDict["Table"] = AF.gridsearch({"FilterType":["LMS"], "NTaps": [Taps], "Mu":MuList}, TrainData, TrainData, Plot=True)
   LMSDict["Filter"] = AF.LMS(**LMSDict["BestPar"])
   LMSDict["MSE"] = LMSDict["Filter"].adapt(TrainData, TrainData, Plot = False, Future=1)[2] # adapt and plot best Filter, keep MSE only
   
   LMSDict["delay"] = int(Taps/5)
   LMSDict["Prediction"] = LMSDict["Filter"].filter(TestData)[LMSDict["delay"]:] # somehow delay compensation of 2 samples is necessary
   LMSDict["PredError"] = np.abs(LMSDict["Prediction"] - TestData[:SigLen-LMSDict["delay"]]) # data needs also to be shortened
   print("With a delay compensation of :", LMSDict["delay"],"prediction error is", LMSDict["PredError"].sum(), "\nThe weights are:", np.round(LMSDict["Filter"].W(), 5) )

   # KLMS 20 for both
   MuList = list(np.linspace(0.01, 2, 1)) # range working well for all noise intensities
   KernelSizeList = list(np.linspace(0.01, 20, 1))
   MinDistList = [1e-07, 5e-07, 1e-06, 1e-05, 1e-04, 1e-03]
   MinErrList  = [1e-06, 1e-05, 5e-05, 1e-04, 5e-04]

   KLMSDict["BestPar"], LMSDict["Table"] = AF.gridsearch({"FilterType":["KLMS"], "Mu": MuList, "KernelSize": KernelSizeList,
                                                          "WindowSize": [Taps], "NC": [True], "NC_MinDist": MinDistList, "NC_MinErr": MinErrList },
                                                          TrainData, TrainData, Plot=False, TestData=TestData)
   KLMSDict["Filter"] = AF.KLMS(**KLMSDict["BestPar"])
   KLMSDict["MSE"] = KLMSDict["Filter"].adapt(TrainData, TrainData, Plot = True, Future=1)[2] # adapt and plot best Filter, keep MSE only

   KLMSDict["Prediction"] = KLMSDict["Filter"].filter(TestData)
   KLMSDict["PredError"] = np.abs(KLMSDict["Prediction"] - TestData) 
   print("KMLS prediction error is", KLMSDict["PredError"].sum())
   KLMSDict["Filter"].print()


   # Plotting the results
   Iter = np.arange(len(TestData))
   Fig, Ax = plt.subplots(2,1)
   Fig.suptitle(str(Taps) + " Taps Prediction")
   Ax[0].plot(LMSDict["Prediction"])
   Ax[0].plot(Iter, KLMSDict["Prediction"], Iter, TestData)
   AF.PltFormat(Fig, Ax[0], "Predictions", "Time in Samples", [0, len(Data["KLMS"]["D_"])], "Amplitude", [], ["LMS Prediction", "KLMS Prediction", "Data"], Show = False)
   Ax[1].plot(LMSDict["PredError"]) #plottet separately having different lenghts
   Ax[1].plot(KLMSDict["PredError"])
   AF.PltFormat(Fig, Ax[1], "Prediction error", "Time in Samples", [0, len(Data["KLMS"]["D_"])], "Absolute Error", [], [], Show = False)


## store the correct data
#File = "C:/Users/Stéphane/BCP, A/00 TU/Digital Signal Processing/00 Modules/Adaptive Filter/my project/Code/Prediction" + str(Taps) + "TData.pkl"

#with open(File, "wb+") as F: # store current dict content
#   pickle.dump(PredictResuts, F)
#   F.close()

# ----------------------------------------------------------------------------- show() ------------------------------------------------------------------
plt.show() # display all plots when code is done running